# -*- coding: iso-8859-1 -*-
'''
main.py -- actual implementation of macros in the package.
----
Copyright (C) 2007 Zoran Isailovski
Licensed under the Apache License, Version 2.0 (see http://www.apache.org/licenses/LICENSE-2.0)
'''

DEBUG = True

INHIBIT_ANONYMOUS_RATINGS = False
SHOW_ATTENDANCE_IN_FORM = False
DEFAULT_SORT_ORDER = 'subject' # one of: page, subject, rank, attendance

##INHIBIT_ANONYMOUS_RATINGS = True
##SHOW_ATTENDANCE_IN_FORM = True
##DEFAULT_SORT_ORDER = 'rank'

######################################################################
## 
## Moin Interface
## 
######################################################################

Dependencies = ["pages"] # footnote: should not be cached
#Dependencies = ['time'] # footnote: should not be cached

def execute(Run, macro, args):
    try:
        return Run(macro, args or '')
    except:
        import sys, traceback
        output = traceback.format_exception( *sys.exc_info() )
        if not DEBUG: output = output[-1:]
        output = ''.join( output )
        return ERROR_FORMAT % (Run.__name__, args, output)

ERROR_FORMAT = '''\
<div class="traceback" style="background-color:red; color:lightyellow; font-weight: bold">
<center>%s(%s)</center>
<pre class="traceback" style="background-color:yellow; color:red; font: 11px verdana; border: solid red">\
%s\
</pre>
</div>
'''

######################################################################
## 
## Implementation
## 
######################################################################
import os

import tmpl
if DEBUG: reload(tmpl)

import datalayer
if DEBUG: reload(datalayer)

def DataModelFactory(attachmentsDir, pagename, subject):
    ##return datalayer.PerSubjectSavedMiniModel(attachmentsDir, subject)
    return datalayer.PerSubjectPickledModel(attachmentsDir, subject)

import MoinMoin.user

######################################################################
######################################################################
######################################################################
## 
## Individual Rating Form
## 
######################################################################
##from core import rating

def DoForm(macro, args):
    InitializeConstants(macro)

    thisPage = macro.formatter.page
    attachmentsDir = thisPage.getPagePath('attachments')
    pagename = thisPage.page_name
    subject = args
    varname = subject + '_var'

    datamodel = DataModelFactory(attachmentsDir, pagename, subject)

    user = macro.request.user
    anonymous = INHIBIT_ANONYMOUS_RATINGS and (not user.valid or not user.exists())
    haveRatingPermission = not anonymous and user.may.write(pagename)
##    username = anonymous and user.auth_username or ''
    username = MoinMoin.user.getUserIdentification(macro.request)

##    # DBG
##    w = macro.request.write 
##    aliasname = getattr(user, 'aliasname', '((no attribute))')
##    d = dict( user=user.name, aliasname=aliasname, trusted=user.trusted, valid=user.valid, exists=user.exists() )
##    s = '<u>user:</u>\n  ' + '\n  '.join([ '%s = %s' % each for each in d.items() ])
##    w( '<pre>%s</pre>' % s)

    thisForm = macro.form
    if thisForm.has_key(varname):
        newRating = int(thisForm[varname][0])
        datamodel.Rate(pagename, subject, username, newRating)

    rankinfo = datamodel.GetRankInfo(pagename, subject, username)
    context = dict(
        subject = subject,
        rank = rankinfo.average,
        myvote = rankinfo.mine,
        nvotes = rankinfo.nvotes,
        anonymous = anonymous,
        haveRatingPermission = haveRatingPermission,
        url = thisPage.url(macro.request),
        varname = varname,
        symbols = Symbols(rankinfo.average, 5),
        showAttendance = SHOW_ATTENDANCE_IN_FORM,
        )
    # helpers for use from within the template
    helpers = dict(
        f2s = lambda x: '%.3g' % x,
        )
    return tmpl.renderThis('form', ctx=context, helpers=helpers)


def Symbols(value, limit=5):
    oncnt = int(round(value))
    urls = OnOffSeq(IMGURL_ON, IMGURL_OFF, oncnt, limit)
    symbols = [ dict(url=v, value=i+1) for (i, v) in enumerate(urls) ]
    return symbols

def OnOffSeq(onsym, offsym, oncnt, limit=5):
    assert 0 <= oncnt <= limit, 'value %r out of range [0..%d]' % (oncnt, limit)
    return [onsym]*oncnt + [offsym]*(limit - oncnt)


######################################################################
## 
## Constants
##
## These are constant throughout a wiki session (and mostly throughout a wiki lifetime)
## but may differ from wiki to wiki.
## 
######################################################################
import cStringIO as StringIO
from MoinMoin.parser import wiki

SYM_ON = '{*}'
SYM_OFF = '{o}'

IMGURL_ON = None
IMGURL_OFF = None

def InitializeConstants(macro):
    global IMGURL_ON, IMGURL_OFF
    if IMGURL_ON is None:
        text = SYM_ON + ' ' + SYM_OFF
        IMGURL_ON, IMGURL_OFF = GetSrcURLs(macro, text)

def GetSrcURLs(macro, text):
    html = wiki2html(macro, text)
    return re.findall(r'src\s*=\s*"(.+?)"', html)

def wiki2html(macro, text):
    stream = StringIO.StringIO()
    macro.request.redirect(stream)
    parser =  wiki.Parser(text, macro.request)
    parser.format(macro.formatter)
    html = stream.getvalue()
    macro.request.redirect()
    stream.close(); del stream
    return html


######################################################################
######################################################################
######################################################################
## 
## Ratings Overview
## 
######################################################################
import re
import glob
import fnmatch

from MoinMoin.Page import Page

def DoOverview(macro, args):
    '''
    moin syntax:
        DoOverview(key [, subject_pattern] [, page_title_pattern])
    '''
    session = Session(macro.request)
    allRatings = []

    # get macro arguments
    args = ParseArgsForOverview(args)

    # analyze html form fields
    thispage = macro.formatter.page
    sortby = macro.form.get('ratings:sortby', [args.sortby])[0]
    try:
        sortby = getattr(SortBy, sortby)
    except AttributeError, SyntaxError:
        sortby = DEFAULT_SORTER

    # collect data into a "table" (i.e. a sequence of records / structs)
    rx_pagename = re.compile( fnmatch.translate(args.pagename) )
    def HasRatings(pagename):
        if not rx_pagename.match(pagename): return False
        ratingsOnThisPage = session.RatingsOnPage(pagename, args.subject)
        allRatings.extend( ratingsOnThisPage )
        return len(ratingsOnThisPage) > 0
    pagenames = macro.request.rootpage.getPageList(filter=HasRatings)

    # sort the table as requested
    allRatings.sort(sortby)

    # provide helpers for use from within the template
    helpers = dict(
        f2s = lambda x: '%.2f' % x,
        succ = lambda n: n + 1
        )

    # render and return the output
    return tmpl.renderThis(
        'overview-table',
        args = args,
        ratings = allRatings,
        thispage = thispage,
        helpers = helpers,
        request = macro.request,
        )

def ParseArgsForOverview(args):
    args = args or ''
    args = re.split(r'\s*,\s*', args)
    data = struct(
        key = 'table',
        ##key = u'ce278e231f577345b73bcff81b9c4b96', ## <= md5.new('ratings:table').hexdigest()
        ##key = u'4ea91377', ##hex(abs(hash('ratings:table')))[2:]
        subject = '*',
        pagename = '*',
        sortby = DEFAULT_SORT_ORDER,
        )
    if args: data.key = args.pop(0) or data.key
    if args: data.subject = args.pop(0) or data.subject
    if args: data.pagename = args.pop(0) or data.pagename
    if args:
        s = args.pop(0)
        m = re.match(r'(?:sortby\s*=\s*)?(\w+)$', s)
        data.sortby = m and m.group(1) or data.sortby
    assert not args, 'Too many macro arguments'
    return data

######################################################################
## 
## Comparision Methods for Sorting
## 
######################################################################

class SortBy:
    @staticmethod
    def SubjectAndRank(a, b):
        c = cmp(a.subject, b.subject)
        if c == 0:
            c = -cmp(a.rank, b.rank)
        return c
    @staticmethod
    def RankAndSubject(a, b):
        c = -cmp(a.rank, b.rank)
        if c == 0:
            c = cmp(a.subject, b.subject)
        return c
    @staticmethod
    def RankAndPage(a, b):
        c = -cmp(a.rank, b.rank)
        if c == 0:
            c = cmp(a.pagename, b.pagename)
        return c
    @staticmethod
    def PageSubjectAndRank(a, b):
        c = cmp(a.pagename, b.pagename)
        if c == 0:
            c = SortBy.SubjectAndRank(a, b)
        return c
    @staticmethod
    def AttendanceSubjectAndRank(a, b):
        c = -cmp(a.nvotes, b.nvotes)
        if c == 0:
            c = SortBy.SubjectAndRank(a, b)
        return c

    page = PageSubjectAndRank
    subject = SubjectAndRank
    rank = RankAndSubject
    attendance = AttendanceSubjectAndRank

DEFAULT_SORTER = getattr(SortBy, DEFAULT_SORT_ORDER)

######################################################################
## 
## Session
## 
######################################################################

class Session:
    def __init__(self, request):
        self.request = request
        self.username = MoinMoin.user.getUserIdentification(request)

    def IterRatingsOnPage(self, pagename, subject='*', showUnrated=True):
        page = Page(self.request, pagename)
        attachmentsDir = page.getPagePath('attachments')
        patterns = re.split(r'\s*\|\s*', subject)
        patterns = [ '%s' % fnmatch.translate(each)[:-1] for each in patterns ]
        pattern = '|'.join(patterns)
        pattern = r'^[^\#].*?\[\[Rating\((%s)\)\]\]' % pattern

        text = page.get_raw_body() or '' # DO NOT USE page.getPageText - it's BUGGY !!!
        subjectsHere = re.findall(pattern, text, re.M)
        if not subjectsHere: return

        for each in subjectsHere:
            datamodel = DataModelFactory(attachmentsDir, pagename, each)
            rankinfo = datamodel.GetRankInfo(pagename, each, self.username)
            rank = rankinfo.average
            if not rank and not showUnrated: continue
            yield struct( pagename=pagename, subject=each, rank=rank, nvotes=rankinfo.nvotes )

    def RatingsOnPage(self, pagename, subject='*'):
        return list( self.IterRatingsOnPage(pagename, subject) )

######################################################################
## 
## Aux
## 
######################################################################

class struct:
    def __init__(self, **fields):
        vars(self).update(fields)
    def __repr__(self):
        return 'struct(**%r)' % vars(self)
